#!/usr/bin/env csi -s
;; eldritch-common.scm
;; Created 2019-04-18
#<<DOC
## BSD License

Copyright © 2019 by Mark Damon Hughes. All Rights Reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
3. Neither the name of the copyright holder nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUDOCENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
DOC

;; Imported source, not a module, use (include-relative "eldritch-common.scm")

;;______________________________________
;; Item type

(define kItem_Loot 'loot)
(define kItem_Light 'light)
(define kItem_Weapon 'weapon)
(define kItem_Armor 'armor)

;; Item type is kItem_* above
;; effect is a score for weapons & armor
;; Most items should weigh 10.
;; (define-record Item name desc type effect score weight)
(define-record-type Item #t #t
	(name)
	(desc)
	(type)
	(effect)
	(score)
	(weight)
)
(define-reader-ctor 'Item make-Item)
(define-record-printer (Item i out)
	(format out "#,(Item ~S ~S ~S ~S ~S ~S)"
		(Item-name i) (Item-desc i) (Item-type i) (Item-effect i) (Item-score i) (Item-weight i)
))
; module exports:
; make-Item Item? Item-name Item-desc Item-type Item-effect Item-score Item-weight

;;______________________________________
;; Doorway type

"proc: symbol name of a door-specific exit function. Must be (lambda () …)."
;; (define-record Doorway dir to-id desc proc)
(define-record-type Doorway #t #t
	(dir)
	(to-id)
	(desc)
	(proc)
)
(define-reader-ctor 'Doorway make-Doorway)
(define-record-printer (Doorway d out)
	(format out "#,(Doorway ~S ~S ~S ~S)"
		(Doorway-dir d) (Doorway-to-id d) (Doorway-desc d) (Doorway-proc d)
))
; module exports:
; make-Doorway Doorway? Doorway-dir Doorway-to-id Doorway-desc Doorway-proc

;;______________________________________
;; Room type

"parse: symbol name of a room-specific command parser. Must be (lambda (cmd line-rest proc) …) where proc is regular command parser. Return #t to use a turn."
;; (define-record Room id name desc lit? exits visited? items parse)
(define-record-type Room #t #t
	(id)
	(name)
	(desc)
	(lit?)
	(exits)
	(visited?)
	(items)
	(parse)
)
(define-reader-ctor 'Room make-Room)
(define-record-printer (Room r out)
	(format out "#,(Room ~S ~S ~S ~S ~S ~S ~S ~S)"
		(Room-id r) (Room-name r) (Room-desc r) (Room-lit? r) (Room-exits r) (Room-visited? r) (Room-items r) (Room-parse r)
))
; module exports:
; make-Room Room? Room-id Room-name Room-desc Room-lit? Room-exits Room-visited? Room-items Room-parse

;; Creates and returns a basic Room.
(define (Room-init id name)
	(make-Room id name "" #t (make-vector kDirCount #f) #f '() #f)
)

;; Adds an exit in door-dir to room.
(define (Room-addexit! room door)
	(vector-set! (Room-exits room) (Doorway-dir door) door)
)

;; Adds room to world.
(define (add-room room)
	(hput! gRooms (Room-id room) room)
)

;;______________________________________
;; Actor type

(define kStatCount 3)
(define kStatNames '("Health" "Sanity" "Luck"))
(define kStatAbbvs '("HEA" "SAN" "LCK"))
(define kStat_HEA 0)
(define kStat_SAN 1)
(define kStat_LCK 2)

;; (define-record Actor name loc stats statsMax score items flags)
(define-record-type Actor #t #t
	(name)
	(loc)
	(stats Actor-stats)
	(statsMax Actor-statsMax)
	(score)
	(items)
	(flags)
)
;; Chicken can't serialize hash-tables, so convert them to alist on save, use custom decoder here on restore.
(define (make-Actor-decode name loc stats statsMax score items flags)
	(make-Actor name loc stats statsMax score items (alist->hash-table flags))
)
(define-reader-ctor 'Actor make-Actor-decode)
(define-record-printer (Actor a out)
	(format out "#,(Actor ~S ~S ~S ~S ~S ~S ~S)"
		(Actor-name a) (Actor-loc a) (Actor-stats a) (Actor-statsMax a) (Actor-score a) (Actor-items a) (hash-table->alist (Actor-flags a))
))
; module exports:
; make-Actor Actor? Actor-name Actor-loc Actor-stats Actor-statsMax Actor-score Actor-items Actor-flags

;; Creates and returns a blank Actor
(define (Actor-init name)
	{let [(pc (make-Actor name #f (make-vector kStatCount 0) (make-vector kStatCount 0) 0 '() (hash-symbol) ))]
		{do [(st 0 (add1 st))] [(>= st kStatCount)]
			{let [(n (* (dice 3 6) 5))]
				(vset! (Actor-stats pc) st n)
				(vset! (Actor-statsMax pc) st n)
			}
		}
		pc
	}
)

(define (Actor-description pc)
	(with-output-to-string (lambda ()
		(printf "~A\nName: ~A\n" kHLine (Actor-name pc))
		{do [(st 0 (add1 st))] [(>= st kStatCount)]
			(printf "~A: ~A [~A]~A"  (pad (anyitem kStatNames st) 6)  (padnum (vget (Actor-stats pc) st) 2)
				(padnum (vget (Actor-statsMax pc) st) 2)  (if (< st (sub1 kStatCount)) "    " "\n")
			)
		}
		(when (notnull? (Actor-items pc))
			(printf "You are carrying:\n")
		)
		(for-each (lambda (it)
				(printf "~A\n" (Item-name it))
			)
			(Actor-items pc)
		)
		(printf "~A\n" kHLine)
	))
)

(define (Actor-weight-carried pc)
	{let [(t 0)]
		(for-each (lambda (it) (inc! t (Item-weight it))) (Actor-items pc))
		t
	}
)

(define (Actor-weight-possible pc)
	(vget (Actor-statsMax pc) kStat_HEA)
)
