#!/usr/bin/env csi -s
;; eldritch.scm
;; Created 2019-04-18
#<<DOC
## BSD License

Copyright © 2019 by Mark Damon Hughes. All Rights Reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
3. Neither the name of the copyright holder nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUDOCENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
DOC

(import scheme)
(import (chicken base))
(import (chicken condition))
(import (chicken format))
(import (chicken io))
(import (chicken pathname))
(import (chicken platform))
(import (chicken port))
(import (chicken process-context))
(import (chicken read-syntax))
(import (chicken string))
;; eggs
(import miscmacros)
(import srfi-1) ;; lists
(import srfi-13) ;; strings
(import srfi-69) ;; NICE. hash-tables
(import srfi-99) ;; records
(import srfi-133) ;; vectors
(import utf8)

(cond-expand
	(compiling (declare (uses marklib)))
	(else (load (make-pathname (pathname-directory (program-name)) "marklib.scm")))
)
(import marklib)

(include-relative "eldritch-common.scm")
(include-relative "eldritch-world.scm")

;;______________________________________
;; Constants

(define kAppTitle "Eldritch")
(define kAppVersion "0.1.2.20190429")
(define kAppURL "https://mdhughes.tech/eldritch/")
(define kAppCopyright "Copyright (c) 2019 Mark Damon Hughes. All Rights Reserved.")
(define kAppCopyrightExtra #<<EOL
EOL
)
(define kUsageText "Usage: mainmenu [-d|--debug] [-v|--verbose]")

(define kHLine "____________________________________________________________________________")

(define kErrorNoRoom "I don't know where you are!")
(define kDataPath "eldritch-data")
(define kSaveExt ".sav")
(define kPlayerStartName "Unknown")

(define gDebug #f)
(define gCommands #f)
(define gRooms #f)
(define gPlayer #f)
(define gTurn 1)

;;______________________________________
;; Player Actions

(define (do-game-about cmd line-rest)
	(msg-plain "~A\n~A ~A\n~A\n~A\n~A" kHLine kAppTitle kAppVersion kAppURL kAppCopyright kAppCopyrightExtra)
	(msg-plain "Built with Chicken Scheme ~A\n" (chicken-version #t))
	(msg-plain "~A~A" (read-text-file (make-pathname kDataPath "about.txt")) kHLine)
	#f
)

(define (do-game-help cmd line-rest)
	(msg-plain "~A\n~A~A" kHLine (read-text-file (make-pathname kDataPath "help.txt")) kHLine)
	#f
)

(define (do-game-name cmd line-rest)
	{let [(newname (string-titlecase (purify-name line-rest)))]
		(cond
			[(string-notempty? newname)
				(set! (Actor-name gPlayer) newname)
				(msg-plain "You set your name to ~A." (Actor-name gPlayer))
			]
			[else  (msg-plain "Your name is ~A." (Actor-name gPlayer))]
		)
	}
	#f
)

(define (do-game-reset cmd line-rest)
	(msg-warning "You will lose your progress and reset the entire world back to start!")
	(when (yesno (prompt-string "Are you sure? "))
		(game-reset)
	)
	#f
)

(define (get-save-filename line-rest)
	{let* [(filename1  (if (string=? line-rest "")  (Actor-name gPlayer)  (purify-name line-rest) ))]
		(string-downcase (if (string=? filename1 "") kPlayerStartName filename1))
	}
)

(define (do-game-save cmd line-rest)
	{let* [(filename (get-save-filename line-rest))]
		(if (write-data-file (make-pathname kDataPath (string-append filename kSaveExt))  (list gPlayer (hash-table-values gRooms)))
			(msg-success "Saved to '~A'!" filename)
			(msg-error "Failed.")
		)
	}
	#f
)

(define (do-game-restore cmd line-rest)
	{let* [(filename (get-save-filename line-rest))
			(data (read-data-file (make-pathname kDataPath (string-append filename kSaveExt))))
			(rooms-list '())
		]
		(cond
			[data
				(set!-values (gPlayer rooms-list) (apply values data))
				(set! gRooms (hash-symbol))
				(for-each add-room rooms-list)
				(msg-success "Restored from '~A'!" filename)
				(do-look "look" "")
				(do-inventory "inv" "")
			]
			[else (msg-error "Failed.")]
		)
	}
	#f
)

(define (do-drop cmd line-rest)
	{let* [(room (player-room))  (item (find-player-item line-rest))]
		(cond
			[(not room)  (msg-error kErrorNoRoom)  #f]
			[(not item)  #f] ; message is given in find-room-item
			[else
				(list-delete! (Actor-items gPlayer) item)
				(list-append! (Room-items room) item)
				(msg-success "You drop ~A." (Item-name item))
				#t
			]
		)
	}
)

(define (do-examine cmd line-rest)
	{let* [(item (find-player-item line-rest))]
		(cond
			[(not item)  #f] ; message is given in find-player-item
			[else (msg-plain "~A: ~A" (Item-name item) (Item-desc item))]
		)
	}
	#f ;; examine should never take time
)

(define (do-inventory cmd line-rest)
	(msg-plain (Actor-description gPlayer))
	#f
)

;; TODO: Look KEYWORD
(define (do-look cmd line-rest)
	{let [(room (player-room))]
		(when gDebug (printf "~A  ~A\n" (Room-id room)  (Room-exits room)))
		(cond
			[(not (Room-lit? room))  (msg-warning "It is dark, you can't see where you are.")]
			[(positive? (string-length line-rest)) ;; look ITEM
				{let* [(item (find-room-item room line-rest))]
					(cond
						[(not item)  #f] ; message is given in find-room-item
						[else (msg-plain "~A: ~A" (Item-name item) (Item-desc item))]
					)
				}
			]
			[else ;; look (room)
				(set! (Room-visited? room) #t)
				(msg-title "~A" (Room-name room))
				(msg-plain "~A" (Room-desc room))
				{do [(dir 0 (add1 dir))] [(>= dir kDirCount)]
					{let [(door (vector-ref (Room-exits room) dir))]
						(when (and door  (string-notempty? (Doorway-desc door))) (msg-plain "~A: ~A" (vget kDirNames (Doorway-dir door)) (Doorway-desc door)))
					}
				}
				(when (notnull? (Room-items room))  (msg-plain "You see here:"))
				(for-each (lambda (it)
						(msg-plain "~A" (Item-name it))
					)
					(Room-items room)
				)
			]
		)
	}
	#f ;; look should never take time
)

(define (do-move cmd line-rest)
	{let [(dir (string->dir cmd))]
		(cond
			[dir  {let* [(room (player-room))  (door (vector-ref (Room-exits room) dir))]
					(cond
						;; TODO: in dark, move randomly?
						[(not door)  (msg-warning "You can't go that way.")  #f]
						[(Doorway-proc door)  (when gDebug (printf "door-proc ~A\n" (Doorway-proc door)))
							((eval (Doorway-proc door)))  #t] ;; call proc
						[else {let [(newloc (Doorway-to-id door))]
							(cond
								[(hget gRooms newloc)  (player-moveto newloc)  #t]
								[else (msg-error "Undefined room '~A'!" newloc)  #f]
							)
						}]
					)
				}
			]
			[else (msg-error "You don't know how to move ~A" cmd)  #f]
		)
	}
)

(define (do-score cmd line-rest)
	(msg-success "Your score is ~A in ~A turns." (player-score)  gTurn)
	#f
)

(define (do-take cmd line-rest)
	{let* [(room (player-room))  (item (find-room-item room line-rest))]
		(cond
			[(not room)  (msg-error kErrorNoRoom)  #f]
			;; TODO: in dark, can't take?
			[(not item)  #f] ; message is given in find-room-item
			[(> (+ (Actor-weight-carried gPlayer) (Item-weight item)) (Actor-weight-possible gPlayer))
				(msg-warning "You are carrying too much.")  #f]
			[else
				(list-delete! (Room-items room) item)
				(list-append! (Actor-items gPlayer) item)
				(msg-success "You take ~A." (Item-name item))
				#t
			]
		)
	}
)

(define (do-admin-teleport cmd line-rest)
	(player-moveto (string->symbol line-rest))
)

;;______________________________________
;; Functions

(define (commands-init)
	(set! gCommands (hash-string))
	(vector-for-each (lambda (a) (hput! gCommands a do-move) ) kDirAbbvs)
	(vector-for-each (lambda (a) (hput! gCommands a do-move) ) kDirNames)
	;; System
	(hput! gCommands "about" do-game-about)
	(hput! gCommands "help" do-game-help)  (hput! gCommands "?" do-game-help)
	(hput! gCommands "name" do-game-name)
	(hput! gCommands "quit" shutdown)
	(hput! gCommands "reset" do-game-reset)
	(hput! gCommands "restore" do-game-restore)
	(hput! gCommands "save" do-game-save)
	;; Actions
	(hput! gCommands "inventory" do-inventory)  (hput! gCommands "inv" do-inventory)  (hput! gCommands "i" do-inventory)
	(hput! gCommands "score" do-score)
	(hput! gCommands "look" do-look)  (hput! gCommands "l" do-look)
	(hput! gCommands "take" do-take)  (hput! gCommands "get" do-take)  (hput! gCommands "t" do-take)
	(hput! gCommands "drop" do-drop)  (hput! gCommands "dr" do-take)
	(hput! gCommands "examine" do-examine)  (hput! gCommands "x" do-examine)
	;; Debugging
	(when gDebug
		(hput! gCommands "teleport" do-admin-teleport)
	)
)

;; Returns the only item that matches 'line-rest' in Actor, #f if not found. Shows an error for no match or more than 1 match.
(define (find-player-item line-rest)
	{let [(matches (filter (lambda (it) (string-contains-ci (Item-name it) line-rest))  (Actor-items gPlayer) ))]
		(cond
			[(null? matches)  (msg-warning "You carry no ~A." line-rest)  #f]
			[(= (length matches) 1)  (car matches)]
			[else (msg-warning "Do you mean: ~A?" (string-intersperse matches ", "))  #f]
		)
	}
)

;; Returns the only item that matches 'line-rest' in room, #f if not found. Shows an error for no match or more than 1 match.
(define (find-room-item room line-rest)
	{let [(matches (filter (lambda (it) (string-contains-ci (Item-name it) line-rest))  (Room-items room) ))]
		(cond
			[(null? matches)  (msg-warning "You see no ~A here." line-rest)  #f]
			[(= (length matches) 1)  (car matches)]
			[else (msg-warning "Do you mean: ~A?" (string-intersperse matches ", "))  #f]
		)
	}
)

(define (game-over)
	(msg-warning "*** GAME OVER ***")
	;; TODO: report score & ranking
	(if (yesno (prompt-string "Do you want to try again? "))
		(game-reset)
		(shutdown)
	)
)

(define (game-reset)
	(world-init)
	(set! gTurn 1)
	(set! gPlayer (Actor-init kPlayerStartName))
	(player-moveto kStartRoomLoc)
)

(define (game-turn)
	(inc! gTurn)
)

(define (player-moveto newloc)
	{let [(room (hget gRooms newloc))]
		(cond
			[room  (set! (Actor-loc gPlayer) newloc)
				(do-look "l" "")
			]
			[else  (msg-error "Invalid room '~A'." newloc)]
		)
	}
)

"Returns current room, or #f if none."
(define (player-room)
	(if gPlayer (hget gRooms (Actor-loc gPlayer)) #f)
)

(define (player-score)
	(+ (Actor-score gPlayer)
		(sum (map Item-score (Actor-items gPlayer)))
		(sum (map Item-score (Room-items (hget gRooms kScoreRoomLoc))))
	)
)

(define (parse line)
	{let* [(words (string-split (string-downcase line)))
			(cmd (if (null? words) #f (car words)))
			(line-rest (if (null? words) "" (string-intersperse (cdr words) " ")))
			(room (player-room))
			(room-parse (if room (Room-parse room)))
			(proc (if cmd (hget gCommands cmd) #f))
		]
		(cond
			[room-parse  (when gDebug (printf "room-parse ~A\n" room-parse))
				((eval room-parse) cmd line-rest proc)]
			[proc  (proc cmd line-rest)]
			[else  (msg-error "I don't know how to '~A'. Try 'help'?" line)  #f]
		)
	}
)

(define (parseargs argv)
	{let argloop [(a (maybe-car argv))  (rest (maybe-cdr argv))]
		(cond
			[(not a)  #f]
			[(or (string=? a "-d") (string=? a "--debug"))  (set! gDebug #t)]
			[else (displayln kUsageText)  (exit 1)]
		)
		(when (notnull? rest)	(argloop (car rest) (cdr rest)))
	}
)

(define (startup)
	;; Low-level setup
	(msg-init)
	(dice-init)
	(commands-init)

	;; reset
	(game-reset)
)

(define (shutdown . args)
	(printf "Bye!\n")
	(exit 0)
)

(define (main argv)
	(parseargs argv)

	;; title
	(clear-screen)
	(msg-title "~A" (read-text-file (make-pathname kDataPath "title.txt")))
	(msg-plain "Type `help` for instructions, `about` for credits.")

	(startup)
	{let loop []
		{let [(line (input (prompt-string "> ")))]
			(when (parse line)
				(game-turn)
			)
		}
		(loop)
	}
)

(main (command-line-arguments))
