;; marklib.scm
;; Created 2018-10-20 14:39:00

#<<DOC

## BSD License

Copyright © 2018,2019 by Mark Damon Hughes. All Rights Reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
3. Neither the name of the copyright holder nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUDOCENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

;;______________________________________

## Introduction

A collection of my utility functions and "how to Scheme" notes.
Anytime I'm likely to re-use something, it gets pulled in here in the appropriate section.

## Style Guide

I'm not interested in/actively hostile to bikeshed arguments about "correct" Scheme style, especially from CLISP users.
This works for me, and makes it easy for me to read and edit.
I've been writing Scheme off and on since the '80s, like this. But I write a lot of other langs, too.

{} around let, do, repeat, until, while, let/cc: Any flow-control block (even if let is just being used for scope).
[] around sub-expressions in cond, let, do.
() around all other expressions.
Prefer `cond` for all multi-line, multi-function choices, use `if` only for ternary choices, `when`/`unless` for single branches.
`begin` is a code smell.
Indent like C with closing paren/brace matched to start if it's more than 1 line.
Indent is tab, displayed as 4 spaces, never mixed with spaces. IT IS AN ERROR to mix tabs and spaces for decoration.
Double space between complex expressions (3 or more parens) in-line, or move to a new line.
lower-case-names for functions, Capitalized for record types, Cap-smallwords for record-related functions, kCapWords for constants, gCapWords for globals.
Docstring above each function, #<<DOC … DOC for multi-line comments. Use Markdown, in particular `code` for parameters. TODO: switch to hahn docstrings.

## Known Problems / Workarounds

- Chicken using SRFI numbers for eggs is… inconvenient.
- Chicken's `format` has minimal placeholders, and the LISP-style `format` egg is non-reentrant, so useless in practice. Use printf or sprintf, ~A for all displays, ~S for debugging, \n for newlines, ~! to flush output, with `pad`, `padnum` defined here to format fields.
- Chicken has no standard docstring system, and recommends putting docs in .egg files. Ha no. Docs further than a few lines from source are always wrong.

## Feedback

If you have *polite* and *code-related* suggestions, you can reach me at @mdhughes@cybre.space on Fediverse.
If you have *impolite* or *bikeshed* suggestions, I will block you.

;;______________________________________

## Snippets:

;; unpack list into variables:
(define-values (x y z) (apply values alist)) (printf "~A,~A,~A\n" x y z)
(set!-values (x y z) (apply values alist)) (printf "~A,~A,~A\n" x y z)
(let-values [[(x y z) (apply values alist)]] (printf "xyz:~A,~A,~A\n" q r s x y z))

;; loops:
{let loop [(i 1)]
	(printf "~A\n" i)
	(when (< i 100)  (loop (add1 i)))
}
{do [(break #f)] [break]
	(set! break #t)
}
{do [(i 0 (add1 i))] [(>= i 10)]
	(printf "~A\n" i)
}
{do [(break #f) (i 0 (add1 i))] [(or break (>= i 10))]
	(printf "~A\n" i)
}
{do [(len (anylen coll)) (i 0 (add1 i))] [(>= i len)]
	(printf "~A: ~A\n" i (anyitem coll i))
}

(define data '(1 2 -3 4))
{call/cc (lambda (break)
	{for-each (lambda (x)
			(printf "~A " x)
			(when (negative? x)  (break x))
		)
		data
	}
	#t
)}

;; from miscmacros, I don't like these as much but they're short:
{dotimes (i 100) (printf "~A\n" i)} ;; counts 0 to 99
{repeat* 100 (printf "~A\n" it)} ;; counts 100 downto 1

;; exception handling
(handle-exceptions ex (begin
		(display "Error: ") (print-error-message ex)
		#f
	)
	;; normal code here
	(error 'location "text" values…)
)

;; For records, write (define-record TYPENAME FIELD...) and then filter with schemeRecordToRecordType.py

DOC

;; FIXME: needs nanosecond timer in C
; (require-extension bind)
; (bind-rename "getTime" "current-microseconds")
; (bind* #<<EOF
; uint64_t getTime();
; #ifndef CHICKEN
; #include <sys/time.h>
; uint64_t getTime() {
;     struct timeval tim;
;     gettimeofday(&tim, NULL);
;     return 1000000 * tim.tv_sec + tim.tv_usec;
; }
; #endif
; EOF
; )

;;______________________________________

(declare (unit marklib))
(module marklib
	(
		;; Logic
		neq? neqv? nequal? notneg? notnull? notzero? !=
		;; Math
		idiv int minmax real sign sum
		;; Strings
		atoi chr numstr ord pad padnum purify-name string-notempty? substr
		;; Data Structures
		anycompare anyeq? anyless? anymore? anylen anyitem
		alget
		hash-symbol hash-number hash-string hadd! hclear hdel! hfill hget hput! hselect hstring
		intlist->string
		list-append! list-delete!
		make-Stack
		maybe-car maybe-cdr
		vget vindex vset!
		;; IO
		read-data-file read-text-file
		write-data-file write-text-file
		;; Random
		dice-init die dice dice-stats
		dice-daro dice-taro
		random-choice rnd
		;; Text UI
		clear-screen
		displayln
		input
		menu
		msg-init msg msg-plain msg-title msg-success msg-warning msg-error Window-Size
		print-table
		prompt-string
		yesno
		;; Time
		iso8601-date iso8601-time
		;; System
		open-url
		;; Geometry
		make-Point Point? Point-x Point-y
		Point-xy-set! Point-x-add! Point-y-add! Point-copy
		Point-line-to-point Point-scale-xy Point-translate

		make-Rect Rect? Rect-x Rect-y Rect-w Rect-h
		Rect-center Rect-contains-point Rect-copy Rect-for-each
		Rect-inset Rect-intersect Rect-intersect? Rect-origin Rect-origin-set! Rect-translate Rect-xy-set!

		kDir_N kDir_NE kDir_E kDir_SE kDir_S kDir_SW kDir_W kDir_NW kDir_U kDir_D kDir_In kDir_Out
		kDirFlatCount kDirCount kDirDelta kDirZ kDirNames kDirAbbvs dir->angle dir-rotate string->dir

		make-Sprite Sprite? Sprite-filename Sprite-sx Sprite-sy Sprite-sw Sprite-sh

		make-Gridmap Gridmap? Gridmap-width Gridmap-height
		Gridmap-init Gridmap-string Gridmap-bounds Gridmap-gridat Gridmap-gridat-set! Gridmap-inbounds? Gridmap-isedge?
	)

(import scheme)
(import (chicken base))
(import (chicken bitwise))
(import (chicken condition))
(import (chicken file))
(import (chicken format))
(import (chicken io))
(import (chicken irregex))
(import (chicken platform))
(import (chicken port))
(import (chicken process))
(import (chicken random))
(import (chicken read-syntax))
(import (chicken sort))
(import (chicken string))
(import (chicken time))
(import (chicken time posix))
;; eggs
(import ioctl)
(import miscmacros)
(import srfi-1) ;; lists
;(import srfi-13) ;; strings
(import srfi-69) ;; NICE. hash-tables
(import srfi-99) ;; records
(import srfi-133) ;; vectors
(import (prefix ansi-escape-sequences "ansi-"))
(import utf8)
(import utf8-srfi-13)
(import utf8-srfi-14)

;;______________________________________
;; Logic

;; Shorthand for (not (comparator …))
(define (neq? a b) (not (eq? a b)))
(define (neqv? a b) (not (eqv? a b)))
(define (nequal? a b) (not (equal? a b)))
(define (notneg? a) (not (negative? a)))
(define (notnull? a) (not (null? a)))
(define (notzero? a) (not (zero? a)))
(define (!= a b) (not (= a b)))

;;______________________________________
;; Math

"Returns `a` divided by `b`, all as integers."
(define (idiv a b)
	(quotient (int a) (int b))
)

"Converts a floating point number `n` to an integer."
(define (int n)
	(if (and (integer? n) (exact? n)) n (inexact->exact (truncate n)) )
)

"Returns `n` clipped to min/max range `a`…`b`"
(define (minmax a b n)
	(max a (min n b))
)

"Converts any number to real."
(define (real n)
	(+ n 0.0)
)

"Returns 1 if `x` > 0, -1 if `x` < 0, else 0."
(define (sign x)
	(cond [(positive? x) 1]
		[(negative? x) -1]
		[else 0]
	)
)

"Adds all numbers in `ls`"
(define (sum ls)
	(foldl + 0 ls)
)

;;______________________________________
;; Strings

"Converts a string to an integer, 0 or `default-value` if invalid.
Strips all leading non-number (0-9, -) elements, and everything from the first trailing non-number to end."
(define (atoi s . default-value)
	{let* [(trimmed (irregex-replace "^([^0-9-]*)(-?[0-9]*)(.*)$" s 2))  (n (string->number trimmed))]
		(if (eqv? n #f)  (if (null? default-value) 0 (car default-value))  (int n) )
	}
)

"(chr n)  Returns Unicode character for number `n`."
(define chr integer->char)

"(ntoa n)  Returns string version of number `n`."
(define numstr number->string)

"Returns the Unicode code point number of `s`, which may be a char or string."
(define (ord s)
	(cond
		[(char? s)  (char->integer s)]
		[(string? s)  (if (positive? (string-length s))  (char->integer (string-ref s 0))  0)]
		[else (error 'ord "Expected char or string but got " s)]
	)
)

"Renders number `n` as string and pads as `pad` below."
(define (padnum n padding)
	(pad (number->string n) padding)
)

"Renders a string with spaces left if `padding` is positive, or right if `padding` is negative, 0 for no padding."
(define (pad s padding)
	{let [(slen (string-length s))]
		(cond [(and (positive? padding) (< slen padding))
				(string-append (make-string (- padding slen) #\space) s)
			]
			[(and (negative? padding) (< slen (- padding)))
				(string-append s (make-string (- (- padding) slen) #\space))
			]
			[else s]
		)
	}
)

"Replaces all non-letter chars with `-`."
(define (purify-name s)
	(irregex-replace/all "-+$" (irregex-replace/all "^-+" (irregex-replace/all "[^A-Za-z-]+" s "-") "") "")
)

"Returns true if `s` is a string with length 1 or more."
(define (string-notempty? s)
	(and (string? s)  (> (string-length s) 0))
)

"A nice substring that clips invalid indices to 0..length."
(define (substr s start end)
	{let [(end-index (string-length s))]
		(substring s (minmax 0 end-index start) (minmax start end-index end))
	}
)

;;______________________________________
;; Data Structures

"Compares any two atoms, returns -1, 0, 1 by ordered comparison, or #f if mismatched or non-atomic types"
(define (anycompare a b)
	(cond
		[(and (number? a) (number? b))  (if (< a b) -1 (if (> a b) 1 0))]
		[(and (string? a) (string? b))  (if (string<? a b) -1 (if (string>? a b) 1 0))]
		[(and (symbol? a) (symbol? b))  (anycompare (symbol->string a) (symbol->string b))]
		[(and (char? a) (char? b))  (if (char<? a b) -1 (if (char>? a b) 1 0))]
		[(and (boolean? a) (boolean? b))  (if (eq? a b) 0 (if (eq? a #f) -1 1))] ;; #f < #t
		[else #f]
	)
)

(define (anyeq? a b)
	(= (anycompare a b) 0)
)

(define (anyless? a b)
	(< (anycompare a b) 0)
)

(define (anymore? a b)
	(> (anycompare a b) 0)
)

"Returns length of any collection `coll` (string, list, vector, hash), #f if unknown type"
(define (anylen coll)
	(cond
		[(list? coll) (length coll)]
		[(string? coll) (string-length coll)]
		[(vector? coll) (vector-length coll)]
		[(hash-table? coll) (hash-table-size coll)]
		[else #f]
	)
)

"Returns subelement `i` of any collection `coll` (string, list, vector, hash), #f if unknown type"
(define (anyitem coll i)
	(cond
		[(and (list? coll) (notnull? coll))		(list-ref coll i)]
		[(string? coll) (string-ref coll i)]
		[(vector? coll) (vector-ref coll i)]
		[(hash-table? coll) (hash-table-ref/default coll i #f)]
		[else #f]
	)
)

"Returns value for symbol `key` in association list `alist`, #f if not found, '() if 0 values, single atom if 1 value, list if 2+ values."
(define (alget alist key)
	(if (and (list? alist)  (positive? (length alist))  (pair? (car alist))) ;; is it an alist?
		{let [(v (assq key alist))]
			;;(printf "(alget alist:~A key:~A value:~A)\n" alist key v)
			(cond
				[(or (not v) (null? v))  #f]				;; not found
				[(and (pair? v) (not (list? v)))  (cdr v)]	;; (key . value)
				[(= (length v) 1)  '()]						;; (key)
				[(= (length v) 2)  (cadr v)]				;; (key value)
				[else  (cdr v)]								;; (key value1 value2...)
			)
		}
		#f
	)
)

"Creates a symbol-keyed hash table"
(define (hash-symbol . args)
	(hfill (make-hash-table eq? symbol-hash 13) args)
)

"Creates a number-keyed hash table"
(define (hash-number . args)
	(hfill (make-hash-table = number-hash 13) args)
)

"Creates a string-keyed hash table"
(define (hash-string . args)
	(hfill (make-hash-table string=? string-hash 13) args)
)

#<<DOC
Adds `n` to current value of `key` in hash table `ht`, default 0.
(hadd! ht key 1) accumulates values in a histogram.
DOC
(define (hadd! ht key n)
	(hput! ht key (+ (hget ht key 0) n))
)

"(hclear ht) Clears a hash table `ht`."
(define hclear hash-table-clear!)

"(hdel! ht key) Deletes a value for `key` from hash table `ht`"
(define hdel! hash-table-delete!)

"Fills a hash-table `ht` with `alist`, every key-value pair is assigned in order."
(define (hfill ht alist)
	(for-each (lambda (p)
			(cond
				[(list? p)  (hput! ht (car p) (cadr p))]  ;; list
				[else  (hput! ht (car p) (cdr p))]  ;; pair
			)
		)
		alist
	)
	ht
)

"Returns a value for `key` from hash table `ht`, or `default` if not found (default of default is #f)."
(define (hget ht key . default-value)
	{let [(default (if (null? default-value) #f (car default-value)))]
		(if (and (hash-table? ht)  key)  (hash-table-ref/default ht key default)  #f)
	}
)

"Sets `value` for `key` in hash table `ht`, deletes if value is #f."
(define (hput! ht key value)
	(if value (hash-table-set! ht key value)
		(hash-table-delete! ht key)
	)
)

"Returns a key in hash table `ht` where the value fulfils `(testfunc value)`, or #f if not found."
(define (hselect ht testfunc)
	{let loop [(keys (hash-table-keys ht))]
		(cond
			[(null? keys)  #f]
			[(testfunc (hget ht (car keys)))  (car keys)]
			[else  (loop (cdr keys))]
		)
	}
)

#<<DOC
Returns a string representation of hash table `ht`,
with optional `prefix` string (default "") before each key,
`delim` string (default ": ") between key and value,
`suffix` string after each value (default "\n").
DOC
(define (hstring ht #!optional prefix delim suffix)
	(with-output-to-string (lambda ()
		(for-each (lambda (key)
				(printf "~A~A~A~A~A" (if prefix prefix "") key (if delim delim ": ") (hash-table-ref ht key) (if suffix suffix "\n"))
			)
			(sort (hash-table-keys ht) anyless?)
		)
	))
)

"Creates a string blob containing all 32-bit integers in `intlist`, byte for byte."
(define (intlist->string intlist)
	{let [(str (make-string (* (length intlist) 4) #\0))	(i 0)]
		(for-each (lambda (x)
				;; big-endian
				(string-set! str i (integer->char (bitwise-and (arithmetic-shift x -24) #xff)))
				(string-set! str (+ i 1) (integer->char (bitwise-and (arithmetic-shift x -16) #xff)))
				(string-set! str (+ i 2) (integer->char (bitwise-and (arithmetic-shift x -8) #xff)))
				(string-set! str (+ i 3) (integer->char (bitwise-and x #xff)))
				(set! i (+ i 4))
			)
			intlist
		)
		str
	}
)

"Puts a new value `value` at the end of a list `ls`"
(define-syntax list-append!
	(syntax-rules ()
		[(_ ls value)	(set! ls (append ls (list value)))]
	)
)

"Deletes value `value` from a list `ls`"
(define-syntax list-delete!
	(syntax-rules ()
		[(_ ls value)	(set! ls (delete value ls))]
	)
)

"Based on TSPL 2.9"
(define (make-Stack)
	(let [(ls '())]
		(lambda (msg . args)
			(case msg
				[(empty?) (null? ls)]
				[(length) (length ls)]
				[(list) ls]
				[(push!) (set! ls (cons (car args) ls))]
				[(peek) (car ls)]
				[(pop!) {let [(x (car ls))] (set! ls (cdr ls)) x}]
				[else (error "invalid Stack command" msg args)]
			)
		)
	)
)

"Returns car of `ls` if it is a list, else #f."
(define (maybe-car ls)
	(if (and (list? ls) (notnull? ls)) (car ls) #f)
)

"Returns cdr of `ls` if it is a list, else null."
(define (maybe-cdr ls)
	(if (and (list? ls) (notnull? ls)) (cdr ls) '())
)

"(vget v i)  Returns `i` index of vector `v`, or #f if out of bounds."
(define (vget vec i)
	(if (and (vector? vec)  (>= i 0)  (< i (vector-length vec)))
		(vector-ref vec i)
		#f
	)
)

"Returns index of `x` (compared with equal?) in vector `vec`, or #f if not found."
(define (vindex vec x)
	(vector-index (lambda (y) (equal? x y)) vec)
)

"(vset! v i x)  Sets `i` index of vector `v` to `x`."
(define vset! vector-set!)

;;______________________________________
;; IO

"Reads a single datum from `filename` and closes it, returns datum or #f if an error occurred."
(define (read-data-file filename)
	(handle-exceptions ex
		(begin
			(printf "Error reading data file ~A: " filename)
			(print-error-message ex)
			#f
		)
		{let* [(port (open-input-file filename)) (data (read port))]
			(close-input-port port)
			data
		}
	)
)

"Reads entire text contents from `filename` and closes it, returns text or #f if an error occurred."
(define (read-text-file filename)
	(handle-exceptions ex
		(begin
			(printf "Error reading text file ~A: " filename)
			(print-error-message ex)
			#f
		)
		{let* [(port (open-input-file filename)) (text (read-string #f port))]
			(close-input-port port)
			text
		}
	)
)

"Writes a single datum `data` to `filename` and closes it, returns #t, or #f if an error occurred."
(define (write-data-file filename data)
	(handle-exceptions ex
		(begin
			(printf "Error writing data file ~A: " filename)
			(print-error-message ex)
			#f
		)
		{let [(port (open-output-file filename))]
			(write data port)
			(close-output-port port)
			#t
		}
	)
)

"Writes a string `text` to `filename` and closes it, returns #t, or #f if an error occurred."
(define (write-text-file filename text)
	(handle-exceptions ex
		(begin
			(printf "Error writing text file ~A: " filename)
			(print-error-message ex)
			#f
		)
		{let [(port (open-output-file filename))]
			(write-string text #f port)
			(close-output-port port)
			#t
		}
	)
)

;;______________________________________
;; Random
;; "Anyone who attempts to generate random numbers by deterministic means is,
;; of course, living in a state of sin." —John von Neumann

"Call before using dice. If an arg is given, it should be a blob or string of 128 bytes."
(define (dice-init #!optional seed)
	;; used to call randomize on default, still could but random's startup should use system-random.
	(when seed (set-pseudo-random-seed! seed) )
)

"Rolls 1 `s`-sided die"
(define (die s)
	(add1 (pseudo-random-integer s))
)

"Rolls `n` `s`-sided dice."
(define (dice n s)
	(cond
		[(= n 1) (die s)]
		[else (+ (die s) (dice (sub1 n) s))]
	)
)

#<<DOC
Generates a vector table of `count` die rolls of `n` d `s` to count of results.
Use (print-table tbl) to display results.
DOC
(define (dice-stats n s count)
	{let [(tbl (make-vector (* n (add1 s)) 0))]
		{repeat count
			{let [(r (dice n s))]
				(vector-set! tbl r (add1 (vector-ref tbl r)))
			}
		}
		tbl
	}
)

"Rolls 2d6, doubles add and roll over. Pass `out` to print the die roll, #t for stdout, #f to be silent."
(define (dice-daro out)
	{let [(a (die 6))  (b (die 6))]
		(when out (format out "(~A,~A)" a b))
		(cond
			[(!= a b) (+ a b)]
			[else (+ a b (dice-daro out))]
		)
	}
)

"Rolls 3d6, triples add and roll over. Pass `out` to print the die roll, #t for stdout, #f to be silent."
(define (dice-taro out)
	{let [(a (die 6))  (b (die 6))  (c (die 6))]
		(when out (format out "(~A,~A,~A)" a b c))
		(cond
			[(not (= a b c)) (+ a b c)]
			[else (+ a b c (dice-taro out))]
		)
	}
)

"Returns a random item from sequence `ls`."
(define (random-choice ls)
	(anyitem ls (sub1 (die (anylen ls))))
)

"Returns a random real number from 0…<1"
(define (rnd)
	(pseudo-random-real)
)

;;______________________________________
;; Text UI

;; Known window size as a Point, will be set in startup if possible
(define Window-Size #f)

"Call before using any `msg` function."
(define (msg-init)
	{let [(winsize (ioctl-winsize))]
		(if winsize
			(set! Window-Size (make-Point (cadr winsize) (car winsize)))  ;; ioctl reports HEIGHT WIDTH
			(set! Window-Size (make-Point 80 25))  ;; output stream, probably 80x25
		)
	}
)

"Clears screen with white-on-black default, and homes cursor with ANSI"
(define (clear-screen)
	(display (ansi-set-text '(bg-black fg-white) "" #f)) ;; no reset
	(display (ansi-cursor-position 1 1))
	(display (ansi-erase-display))
	(flush-output)
)

"Convenience to display & newline a single object."
(define (displayln obj)
	(display obj) (newline) (flush-output)
)

"Displays prompt if not #f, reads a line from stdin, ends program on EOF"
(define (input prompt)
	(when prompt (printf "~A~!" prompt))
	{let [(s (read-line))]
		(cond
			[(eof-object? s) (printf "Bye!\n") (exit)]
			[else s]
		)
	}
)

#<<DOC
Displays a 1-indexed menu of `items`, and an input `prompt`.
Returns item chosen (0-indexed), or -1 if `cancel` is true, repeats prompt if invalid.
Example: (menu "Letter? " '("A" "B" "C") #t)
DOC
(define (menu prompt items cancel)
	(newline)
	(when cancel (printf "0) Cancel\n"))
	{do [(i 0 (add1 i))] [(>= i (anylen items))]
		(printf "~A) ~A\n" (add1 i) (anyitem items i))
	}
	{let loop [(sel -1)]
		(set! sel (atoi (input prompt) -1))
		(cond
			[(and (>= sel (if cancel 0 1)) (<= sel (anylen items)))  (sub1 sel)]
			[else (loop sel)]
		)
	}
)

#<<DOC
Takes a format pattern and args, word-wraps to window, and paginates long text.
Always adds a trailing newline.
DOC
(define (msg text . args)
	;; TODO: word-wrap at window, count lines, and paginate
	(apply printf text args)
	(newline)
)

"Calls `msg` with white-on-black text."
(define (msg-plain text . args)
	(apply msg (ansi-set-text '(bg-black fg-white) text) args)
)

"Calls `msg` with yellow-on-black text."
(define (msg-title text . args)
	(apply msg (ansi-set-text '(bg-black fg-yellow bold) text) args)
)

"Calls `msg` with green-on-black text."
(define (msg-success text . args)
	(apply msg (ansi-set-text '(bg-black fg-green) text) args)
)

"Calls `msg` with magenta-on-black text."
(define (msg-warning text . args)
	(apply msg (ansi-set-text '(bg-black fg-magenta) text) args)
)

"Calls `msg` with red-on-black text."
(define (msg-error text . args)
	(apply msg (ansi-set-text '(bg-black fg-red) text) args)
)

"Prints a tab-delimited table of index, value"
(define (print-table tbl)
	{do [(len (anylen tbl)) (i 1 (add1 i))] [(>= i len)]
		(printf "~A\t~A\n" i (anyitem tbl i))
	}
)

"Returns `s` colored cyan for a user prompt."
(define (prompt-string s)
	(ansi-set-text '(bg-black fg-cyan) s #f)
)

"Displays `prompt` and waits for user to type a line starting y or n, case-insensitive. Returns #t or #f"
(define (yesno prompt)
	{let loop [(ans "")]
		(set! ans (substr (input prompt) 0 1))
		(cond [(string-ci=? ans "y") #t]
			[(string-ci=? ans "n") #f]
			[else (msg-warning "Answer Yes or No.") (loop ans)]
		)
	}
)

;;______________________________________
;; Time

"Returns ISO8601-formatted date for Unix epoch `seconds`, or current if #f"
(define (iso8601-date seconds)
	(time->string (seconds->utc-time (if seconds seconds (current-seconds))) "%Y-%m-%d")
)

"Returns ISO8601-formatted time for Unix epoch `seconds`, or current if #f"
(define (iso8601-time seconds)
	(time->string (seconds->utc-time (if seconds seconds (current-seconds))) "%H:%M:%S")
)

;;______________________________________
;; System

(define (open-url url)
	(case (software-version)
		[(macosx)  (system (sprintf "open \"~A\"" url))]
		[else  (printf "WARNING: open-url OS ~A is not supported" (software-version))]
	)
)

;;______________________________________
;; Geometry

"Point type"
;; (define-record Point x y)
(define-record-type Point #t #t
	(x)
	(y)
)
(define-reader-ctor 'Point make-Point)
(define-record-printer (Point p out)
	(format out "#,(Point ~S ~S)"
		(Point-x p) (Point-y p)
))

(define (Point-x-add! pt n)
	(set! (Point-x pt) (+ (Point-x pt) n))
)

(define (Point-y-add! pt n)
	(set! (Point-y pt) (+ (Point-y pt) n))
)

(define (Point-xy-set! pt x y)
	(set! (Point-x pt) x)
	(set! (Point-y pt) y)
)

(define (Point-copy pt)
	(make-Point (Point-x pt) (Point-y pt))
)

"Returns a list of integer points from `pt1` to `pt2`."
(define (Point-line-to-point pt1 pt2)
	;;(printf "Point-line-to-point pt1:~A pt2:~A\n" pt1 pt2)
	{let* [(delta (make-Point (- (Point-x pt2) (Point-x pt1))  (- (Point-y pt2) (Point-y pt1))))
			(nsteps (+ (abs (Point-x delta)) (abs (Point-y delta))))
			(step (make-Point (/ (real (Point-x delta)) nsteps)  (/ (real (Point-y delta)) nsteps)))
			(points '())
		]
		{do [(d 0 (add1 d))] [(> d nsteps)] ;; ends equal to nsteps
			{let* [(ptreal (make-Point (+ (Point-x pt1) (* d (Point-x step)))  (+ (Point-y pt1) (* d (Point-y step))) ))
					(pti (make-Point (int (Point-x ptreal))  (int (Point-y ptreal)) ))
				]
				;;(printf "~A  ~A\n" ptreal pti)
				(when (or (null? points) (nequal? pti (last points)))  (list-append! points pti) )
			}
		}
		points
	}
)

"Returns a new Point by multiplying `pt` by `x,y`."
(define (Point-scale-xy pt x y)
	(make-Point (* (Point-x pt) x)  (* (Point-y pt) y) )
)

"Returns a new Point adding `pt1` and `pt2` coordinates."
(define (Point-translate pt1 pt2)
	(make-Point (+ (Point-x pt1) (Point-x pt2))  (+ (Point-y pt1) (Point-y pt2)) )
)

;;______________________________________

"Rect type"
;; (define-record Rect x y w h)
(define-record-type Rect #t #t
	(x)
	(y)
	(w)
	(h)
)
(define-reader-ctor 'Rect make-Rect)
(define-record-printer (Rect r out)
	(format out "#,(Rect ~S ~S ~S ~S)"
		(Rect-x r) (Rect-y r) (Rect-w r) (Rect-h r)
))

"Returns integer center point of `rect`"
(define (Rect-center rect)
	(make-Point (+ (Rect-x rect) (idiv (Rect-w rect) 2))  (+ (Rect-y rect) (idiv (Rect-h rect) 2)))
)

"Returns true if `arect` contains point `pt`."
(define (Rect-contains-point arect pt)
	(and (>= (Point-x pt) (Rect-x arect))  (< (Point-x pt) (+ (Rect-x arect) (Rect-w arect)))
			(>= (Point-y pt) (Rect-y arect))  (< (Point-y pt) (+ (Rect-y arect) (Rect-h arect)))
	)
)

(define (Rect-copy rect)
	(make-Rect (Rect-x rect) (Rect-y rect) (Rect-w rect) (Rect-h rect))
)

"Calls (filter pt) for each point in `rect`"
(define (Rect-for-each rect filter)
	{let [(pt (make-Point 0 0))  (x0 (Rect-x rect))  (x1 (+ (Rect-x rect) (Rect-w rect)))  (y0 (Rect-y rect))  (y1 (+ (Rect-y rect) (Rect-h rect)))]
		{do [(y y0 (add1 y))] [(>= y y1)]
			{do [(x x0 (add1 x))] [(>= x x1)]
				(Point-xy-set! pt x y)
				(filter pt)
			}
		}
	}
)

#<<DOC
Returns `rect` inset by left, right, top, buttom.
+delta to shrink, -delta to grow.
DOC
(define (Rect-inset rect dx-left dx-right dy-top dy-bottom)
	(make-Rect (+ (Rect-x rect) dx-left) (+ (Rect-y rect) dy-top) (- (Rect-w rect) (+ dx-left dx-right)) (- (Rect-h rect) dy-top dy-bottom))
)

"Returns a new rect of the intersection between `arect` and `brect`, coords & size 0 if not intersecting."
(define (Rect-intersect arect brect)
	{let [(x (max (Rect-x arect) (Rect-x brect)))  (x2 (min (+ (Rect-x arect) (Rect-w arect)) (+ (Rect-x brect) (Rect-w brect))))
		(y (max (Rect-y arect) (Rect-y brect)))  (y2 (min (+ (Rect-y arect) (Rect-h arect)) (+ (Rect-y brect) (Rect-h brect))))]
		(cond
			[(and (> x2 x) (> y2 y))  (make-Rect x y (- x2 x) (- y2 y))]
			[else (make-Rect 0 0 0 0)]
		)
	}
)

"Returns true if `arect` intersects `brect`."
(define (Rect-intersect? arect brect)
	(not (or (>= (Rect-x arect) (+ (Rect-x brect) (Rect-w brect)))
		(<= (+ (Rect-x arect) (Rect-w arect)) (Rect-x brect))
		(>= (Rect-y arect) (+ (Rect-y brect) (Rect-h brect)))
		(<= (+ (Rect-y arect) (Rect-h arect)) (Rect-y brect))
	))
)

"Returns origin point of `rect`."
(define (Rect-origin rect)
	(make-Point (Rect-x rect) (Rect-y rect))
)

"Sets origin of `rect` to `pt`"
(define (Rect-origin-set! rect pt)
	(set! (Rect-x rect) (Point-x pt))
	(set! (Rect-y rect) (Point-y pt))
)

"Returns `rect` translated by `offset` Point."
(define (Rect-translate rect offset)
	(make-Rect (+ (Rect-x rect) (Point-x offset))  (+ (Rect-y rect) (Point-y offset))  (Rect-w rect)  (Rect-h rect))
)

"Sets origin of `rect` to `x,y`"
(define (Rect-xy-set! rect x y)
	(set! (Rect-x rect) x)
	(set! (Rect-y rect) y)
)

;;______________________________________
;; Direction type

(define kDir_N 0)
(define kDir_NE 1)
(define kDir_E 2)
(define kDir_SE 3)
(define kDir_S 4)
(define kDir_SW 5)
(define kDir_W 6)
(define kDir_NW 7)
(define kDirFlatCount 8) ;; horizontal only
(define kDir_U 8)
(define kDir_D 9)
(define kDir_In 10)
(define kDir_Out 11)
(define kDirCount 12) ;; horizontal and vertical
"XY-coord, assuming 0,0 is top-left."
(define kDirDelta (vector
	(make-Point 0 -1)  ;; kDir_N
	(make-Point 1 -1)  ;; kDir_NE
	(make-Point 1 0)  ;; kDir_E
	(make-Point 1 1)  ;; kDir_SE
	(make-Point 0 1)  ;; kDir_S
	(make-Point -1 1)  ;; kDir_SW
	(make-Point -1 0)  ;; kDir_W
	(make-Point -1 -1)  ;; kDir_NW
	(make-Point 0 0)  ;; kDir_U
	(make-Point 0 0)  ;; kDir_D
	(make-Point 0 0)  ;; kDir_In
	(make-Point 0 0)  ;; kDir_Out
))
"Z-coord, assuming negative is up, positive is down, usually true in dungeon games."
(define kDirZ (vector
	0 0 0 0 0 0 0 0 -1 1 0 0
))
(define kDirNames (vector "north" "northeast" "east" "southeast" "south" "southwest" "west" "northwest" "up" "down" "inside" "outside"))
(define kDirAbbvs (vector "n" "ne" "e" "se" "s" "sw" "w" "nw" "u" "d" "in" "out"))

"Returns direction as an angle (in degrees), counting kDir_N as 0."
(define (dir->angle f)
	(if (< f kDirFlatCount)
		(modulo (* f 45) 360)
		0
	)
)

"Rotates direction on horizontal plane."
(define (dir-rotate f n)
	(if (< f kDirFlatCount)
		(modulo (+ f n kDirFlatCount) kDirFlatCount)
		f
	)
)

"Returns direction for name"
(define (string->dir s)
	{let* [(sl (string-downcase s))]
		(or (vindex kDirAbbvs sl)  (vindex kDirNames sl) )
	}
)

;;______________________________________

"Sprite type: image, source rectangle"
;; (define-record Sprite filename sx sy sw sh)
(define-record-type Sprite #t #t
	(filename)
	(sx)
	(sy)
	(sw)
	(sh)
)
(define-reader-ctor 'Sprite make-Sprite)
(define-record-printer (Sprite s out)
	(format out "#,(Sprite ~S ~S ~S ~S ~S)"
		(Sprite-filename s) (Sprite-sx s) (Sprite-sy s) (Sprite-sw s) (Sprite-sh s)
))

;;______________________________________

"Gridmap type: 2D array, grids may be any type."
;; (define-record Gridmap width height grids)
(define-record-type Gridmap #t #t
	(width)
	(height)
	(grids Gridmap-grids)
)
(define-reader-ctor 'Gridmap make-Gridmap)
(define-record-printer (Gridmap g out)
	(format out "#,(Gridmap ~S ~S ~S)"
		(Gridmap-width g) (Gridmap-height g) (Gridmap-grids g)
))

"Creates an empty Gridmap, filled by calling `(gridfunc gm pt)`"
(define (Gridmap-init width height gridfunc)
	{let [(gm (make-Gridmap width height (make-vector (* width height) #f)))]
		(Rect-for-each (Gridmap-bounds gm) (lambda (pt)
			(Gridmap-gridat-set! gm pt (gridfunc gm pt))
		))
		gm
	}
)

"Creates a string containing all the content of tilemap, optionally passed through `(gridfunc g)`."
(define (Gridmap-string gm #!optional gridfunc)
	(with-output-to-string (lambda ()
		(Rect-for-each (Gridmap-bounds gm) (lambda (pt)
			(when (and (zero? (Point-x pt)) (positive? (Point-y pt))) (newline))
			{let [(g (Gridmap-gridat gm pt))]
				(printf "~A " (if gridfunc (gridfunc g)  g))
			}
		))
		(newline)
	))
)

"Returns a Rect of the bounds of the Gridmap"
(define (Gridmap-bounds gm)
	(make-Rect 0 0 (Gridmap-width gm) (Gridmap-height gm))
)

"Returns grid at `pt` or #f if out of bounds."
(define (Gridmap-gridat gm pt)
	(cond
		[(Gridmap-inbounds? gm pt)
			{let [(i (+ (Point-x pt) (* (Point-y pt) (Gridmap-width gm))))]
				(vector-ref (Gridmap-grids gm) i)
			}
		]
		[else #f]
	)
)

"Sets grid at `pt` to `g` if in bounds."
(define (Gridmap-gridat-set! gm pt g)
	(cond
		[(Gridmap-inbounds? gm pt)
			{let [(i (+ (Point-x pt) (* (Point-y pt) (Gridmap-width gm))))]
				(vector-set! (Gridmap-grids gm) i g)
			}
		]
		[else #f]
	)
)

"Returns #t if `pt` is inside `gm`"
(define (Gridmap-inbounds? gm pt)
	(and (>= (Point-x pt) 0)  (< (Point-x pt) (Gridmap-width gm))
		(>= (Point-y pt) 0)  (< (Point-y pt) (Gridmap-height gm))
	)
)

"Returns #t if `pt` is at an edge of `gm`."
(define (Gridmap-isedge? gm pt)
	(or (= (Point-x pt) 0)  (= (Point-x pt) (sub1 (Gridmap-width gm)))
		(= (Point-y pt) 0)  (= (Point-y pt) (sub1 (Gridmap-height gm)))
	)
)

) ;; end module marklib
